/*
 * Copyright (C) 2011 Koji Suzuki
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, version 2.1 of the License.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

`include "rtavr_defs.v"
// global defines : USE_DMY_CLOCK, PRC_WATCH, GPR_INITCLR

//`define USE_DMY_CLOCK
//`define PRC_WATCH
//`define GPR_INITCLR

// local defines : 
// GPR_TUNE_1 : unuse v_DOBL,v_DOBH : default
`define GPR_TUNE_1
// GPR_TUNE_2 : no FORWARD_TO_BH (dual port +4)
`define GPR_TUNE_2

// GPR modele for the Reduced Core TinyAVR (attiny10) 
// version 005
// 
// Target Device: xc3s200a-4ft256
// Product Version: ISE 12.4
// Design Goal: Balanced
// LATCH_DOL : Insert LATCHes for Timing
//                                      
// Number of Slice Flip Flops:              60
// Number of 4 input LUTs:                 138
// Number of occupied Slices:               84
// Total Number of 4 input LUTs:           139
// Number used for Dual Port RAMs:          16
// Number of bonded IOBs:                   74
// Post-PAR Static Timing Report
// Clock to Setup (CLK2X) MAX          
//
//
module rtavr_gpr_16(
		input RESET,
`ifdef USE_DMY_CLOCK
		input CLK2X_IN,
		output CLK_OUT,
`else
		input CLK2X,
		input CLK,
`endif
		input WE,
		input PREDEC, POSTINC, 
		input [7:0] DI,
		input [3:0] ADDRAL,
		input [3:0] ADDRAH,
		input [3:0] ADDRBL,
		input [3:0] ADDRBH,
		output [7:0] DOBL,
		output [7:0] DOBH
		, output [11:0] JA
		, output [15:0] INDEX
`ifdef PRC_WATCH
		, output [3:0] WB_ADDR
`endif
`ifdef S1_SKIP_GPR
		, input [2:0] RD_BIT
		, output RD_BIT_OUT
		, output MATCH_RDRR
`endif
		);
// Timing MEMO:
//
//                   _______________             ______________
// CLK2X ____________|              |____________|             |
//     __                              ___________________________
// CLK   |____________________________|                          |
//
//
//  --- F ------------------------ F -----------
//        ADDRs
//        PREDEC/POSTINC
//       ----------- F ------------------------- F
//              WE
//                     - - - - - - - - - ------- F ---
//                                         DI
//                   V                           V
//                   FORWARD_cond  
//                   WRITE WB_DATA               WRITE DI or WB_DATA
//
//        
//                                  V                          V
//                   
//                                  s1_DOBL                    DOBL
//                                                             DOBH

`ifdef USE_DMY_CLOCK
  reg CLK;
  wire CLK2X;
  assign CLK_OUT = CLK;
  assign CLK2X   = CLK2X_IN;
`endif
  reg  [3:0] r_addra;
  reg  [3:0] r_addrb;

  reg  [7:0] s2_DOBL;
  reg  [7:0] s2_DOBH;
  reg  [7:0] gpr [0:15];

  assign DOBL = s2_DOBL;
  assign DOBH = s2_DOBH;

  wire [8:0] DOA = { 1'b0, gpr[r_addra] };

  reg  c;
  wire WB = (PREDEC | POSTINC);
  reg  [3:0] r_wb_addr;

  `ifdef PRC_WATCH
    assign WB_ADDR = r_wb_addr;
  `endif

  wire [7:0] DOA2 = (PREDEC & (~CLK | c) ) ? DOA - 1 : DOA;
  wire [8:0] WB_DATA = (PREDEC & (~CLK | c) ) ?  {1'b0 , DOA} - 1
                     : (POSTINC & (~CLK | c)) ?  {1'b0 , DOA} + 1
		     : {1'b0 , DOA};
  
  wire WE2 = WB | (WE & CLK) ;

  reg  [7:0] s1_DOAL;
  reg  [7:0] s1_DOAH;
  reg  [7:0] s1_DOBL;

  reg r_forward_to_bl;
  wire FORWARD_TO_BL = r_forward_to_bl;

  wire [7:0] v_DOBL = FORWARD_TO_BL ? DI : s1_DOBL;

`ifdef GPR_TUNE_2
  wire [7:0] v_DOBH = gpr[r_addrb];
`else
  reg r_forward_to_bh;
  wire FORWARD_TO_BH = r_forward_to_bh;

  wire [7:0] v_DOBH = FORWARD_TO_BH ? DI : gpr[r_addrb];
`endif

`ifdef S1_SKIP
  wire MATCH_RDRR = (v_DOBL == v_DHBH);

  wire RD_BIT_OUT =  (RD_BIT == 0) ? v_DOBH[0]
                   : (RD_BIT == 1) ? v_DOBH[1] 
		   : (RD_BIT == 2) ? v_DOBH[2]
		   : (RD_BIT == 3) ? v_DOBH[3]
		   : (RD_BIT == 4) ? v_DOBH[4]
		   : (RD_BIT == 5) ? v_DOBH[5]
		   : (RD_BIT == 6) ? v_DOBH[6]
		   : v_DOBH[7];
`endif

`ifdef GPR_TUNE_1
    assign JA[7:0] = FORWARD_TO_BL ? DI : s1_DOBL;
    assign INDEX[7:0] = s1_DOAL;

  `ifdef GPR_TUNE_2
    assign JA[11:8] = gpr[r_addrb][3:0];
    assign INDEX[15:8] = (WB) ? s1_DOAH : gpr[r_addrb];
  `else
    assign JA[11:8] = FORWARD_TO_BH ? DI[3:0] : gpr[r_addrb][3:0];
    assign INDEX[15:8] = (WB) ? s1_DOAH : FORWARD_TO_BH ? DI : gpr[r_addrb];
  `endif
`else
    assign JA[7:0] = v_DOBL;
    assign JA[11:8] = v_DOBH[3:0];
    assign INDEX[7:0] = s1_DOAL;
    assign INDEX[15:8] = (WB) ? s1_DOAH : v_DOBH ;
`endif

`ifdef USE_DMY_CLOCK
always @(negedge CLK2X)
  begin
	CLK <= ~CLK;
  end

initial
  begin
	CLK <= 0;
  end
`endif

always @(posedge CLK2X)
  begin
    if (RESET == 0)
      begin
	c <= 0;
	s1_DOAL <= 0;
	s1_DOAH <= 0;
	r_wb_addr <= 0;
      end
    else
      begin
	if (WE2) gpr[r_addra] <= (WB) ? WB_DATA : DI ;
	if (~CLK) s1_DOAL <= DOA2[7:0];
	if (CLK) s1_DOAH <= DOA2[7:0];
	r_wb_addr <= (WE) ? ADDRAH : ADDRBH;
	c <= WB_DATA[8];
      end
  end

always @(negedge CLK2X)
  begin
    if (RESET == 0)
      begin
	r_addra <= 0;
	r_addrb <= 0;
	s1_DOBL <= 0;
	s2_DOBL <= 0;
	s2_DOBH <= 0;

	r_forward_to_bl <= 0;
`ifdef GPR_TUNE_2
`else
	r_forward_to_bh <= 0;
`endif
      end
    else
      begin
	r_addra <= CLK ? ADDRAL : r_wb_addr;
	r_addrb <= CLK ? ADDRBL : ADDRBH;
	if (~CLK) s1_DOBL <= gpr[r_addrb];
	if (CLK) s2_DOBL <= v_DOBL;
	if (CLK) s2_DOBH <= v_DOBH;

	r_forward_to_bl <=  (WE) & (r_wb_addr == ADDRBL);
`ifdef GPR_TUNE_2
`else
	r_forward_to_bh <=  (WE) & (r_wb_addr == ADDRBH);
`endif
      end
  end

`ifdef GPR_INITCLR
  integer i;
  initial
    begin
      for (i = 0 ; i < 16 ; i = i + 1)
	gpr[i] <= 8'h00;
    end
`endif
endmodule
